/*
 * kgraft_patch_bsc1187054
 *
 * Fix for CVE-2021-3573, bsc#1187054
 *
 *  Upstream commits:
 *  e305509e678b (Bluetooth: use correct lock to prevent UAF of hdev object")
 *  e04480920d1e ("Bluetooth: defer cleanup of resources in
 *                 hci_unregister_dev()")
 *
 *  SLE12-SP3 commit:
 *  e08729f111e76adf19d9360a6a4445f46ad41f87
 *
 *  SLE12-SP4, SLE15 and SLE15-SP1 commit:
 *  c8012e0bd7c58a35d6d5b8d3de91ce8a585b6480
 *
 *  SLE12-SP5 commits:
 *  1793fcdbac417c8c821b15a37b6a57e4de160b39
 *  c8012e0bd7c58a35d6d5b8d3de91ce8a585b6480
 *
 *  SLE15-SP2 and -SP3 commit:
 *  6781ea8a7e050563efc951a5ac219e1600d41262
 *  38ad73f7d26b2ba662856a79d998c29dfc554b04
 *
 *
 *  Copyright (c) 2021 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_BT)

#if !IS_MODULE(CONFIG_BT)
#error "Live patch supports only CONFIG_BT=m"
#endif

/* klp-ccp: from net/bluetooth/hci_sock.c */
#include <linux/export.h>
#include <asm/unaligned.h>
#include <net/bluetooth/bluetooth.h>
#include <net/bluetooth/hci_core.h>

/* klp-ccp: from include/net/bluetooth/hci_core.h */
static int (*kgre_hci_dev_open)(__u16 dev);
static int (*kgre_hci_dev_close)(__u16 dev);

static int (*kgre_hci_dev_reset)(__u16 dev);
static int (*kgre_hci_dev_reset_stat)(__u16 dev);
static int (*kgre_hci_dev_cmd)(unsigned int cmd, void __user *arg);
static int (*kgre_hci_get_dev_list)(void __user *arg);
static int (*kgre_hci_get_dev_info)(void __user *arg);
static int (*kgre_hci_get_conn_list)(void __user *arg);
static int (*kgre_hci_get_conn_info)(struct hci_dev *hdev, void __user *arg);
static int (*kgre_hci_get_auth_info)(struct hci_dev *hdev, void __user *arg);
static int (*kgre_hci_inquiry)(void __user *arg);

static int (*kgre_hci_bdaddr_list_add)(struct list_head *list, bdaddr_t *bdaddr, u8 type);
static int (*kgre_hci_bdaddr_list_del)(struct list_head *list, bdaddr_t *bdaddr, u8 type);

/* klp-ccp: from net/bluetooth/hci_sock.c */
#define hci_pi(sk) ((struct hci_pinfo *) sk)

struct hci_pinfo {
	struct bt_sock    bt;
	struct hci_dev    *hdev;
	struct hci_filter filter;
	__u32             cmsg_mask;
	unsigned short    channel;
	unsigned long     flags;
};

static int kgrr_hci_sock_blacklist_add(struct hci_dev *hdev, void __user *arg)
{
	bdaddr_t bdaddr;
	int err;

	if (copy_from_user(&bdaddr, arg, sizeof(bdaddr)))
		return -EFAULT;

	hci_dev_lock(hdev);

	err = (*kgre_hci_bdaddr_list_add)(&hdev->blacklist, &bdaddr, BDADDR_BREDR);

	hci_dev_unlock(hdev);

	return err;
}

static int kgrr_hci_sock_blacklist_del(struct hci_dev *hdev, void __user *arg)
{
	bdaddr_t bdaddr;
	int err;

	if (copy_from_user(&bdaddr, arg, sizeof(bdaddr)))
		return -EFAULT;

	hci_dev_lock(hdev);

	err = (*kgre_hci_bdaddr_list_del)(&hdev->blacklist, &bdaddr, BDADDR_BREDR);

	hci_dev_unlock(hdev);

	return err;
}

/*
 * Fix CVE-2021-3573
 *  -1 line, +1 line
 */
static int __kgrp_hci_sock_bound_ioctl(struct hci_dev *hdev, unsigned int cmd,
				unsigned long arg)
{
	/*
	 * Fix CVE-2021-3573
	 *  -4 lines
	 */

	if (hci_dev_test_flag(hdev, HCI_USER_CHANNEL))
		return -EBUSY;

	if (hci_dev_test_flag(hdev, HCI_UNCONFIGURED))
		return -EOPNOTSUPP;

	if (hdev->dev_type != HCI_BREDR)
		return -EOPNOTSUPP;

	switch (cmd) {
	case HCISETRAW:
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		return -EOPNOTSUPP;

	case HCIGETCONNINFO:
		return (*kgre_hci_get_conn_info)(hdev, (void __user *) arg);

	case HCIGETAUTHINFO:
		return (*kgre_hci_get_auth_info)(hdev, (void __user *) arg);

	case HCIBLOCKADDR:
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		return kgrr_hci_sock_blacklist_add(hdev, (void __user *) arg);

	case HCIUNBLOCKADDR:
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		return kgrr_hci_sock_blacklist_del(hdev, (void __user *) arg);
	}

	return -ENOIOCTLCMD;
}

/*
 * Fix CVE-2021-3573
 *  +24 lines
 */
static int kgrp_hci_sock_bound_ioctl(struct sock *sk, unsigned int cmd,
				unsigned long arg)
{
	int r;
	struct hci_dev *hdev;

	local_bh_disable();
	bh_lock_sock(sk);
	hdev =  hci_pi(sk)->hdev;
	if (!hdev) {
		bh_unlock_sock(sk);
		local_bh_enable();
		return -EBADFD;
	}
	hci_dev_hold(hdev);
	bh_unlock_sock(sk);
	local_bh_enable();

	r = __kgrp_hci_sock_bound_ioctl(hdev, cmd, arg);

	hci_dev_put(hdev);

	return r;
}

int kgrp_hci_sock_ioctl(struct socket *sock, unsigned int cmd,
			  unsigned long arg)
{
	void __user *argp = (void __user *) arg;
	struct sock *sk = sock->sk;
	int err;

	BT_DBG("cmd %x arg %lx", cmd, arg);

	lock_sock(sk);

	if (hci_pi(sk)->channel != HCI_CHANNEL_RAW) {
		err = -EBADFD;
		goto done;
	}

	release_sock(sk);

	switch (cmd) {
	case HCIGETDEVLIST:
		return (*kgre_hci_get_dev_list)(argp);

	case HCIGETDEVINFO:
		return (*kgre_hci_get_dev_info)(argp);

	case HCIGETCONNLIST:
		return (*kgre_hci_get_conn_list)(argp);

	case HCIDEVUP:
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		return (*kgre_hci_dev_open)(arg);

	case HCIDEVDOWN:
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		return (*kgre_hci_dev_close)(arg);

	case HCIDEVRESET:
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		return (*kgre_hci_dev_reset)(arg);

	case HCIDEVRESTAT:
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		return (*kgre_hci_dev_reset_stat)(arg);

	case HCISETSCAN:
	case HCISETAUTH:
	case HCISETENCRYPT:
	case HCISETPTYPE:
	case HCISETLINKPOL:
	case HCISETLINKMODE:
	case HCISETACLMTU:
	case HCISETSCOMTU:
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		return (*kgre_hci_dev_cmd)(cmd, argp);

	case HCIINQUIRY:
		return (*kgre_hci_inquiry)(argp);
	}

	lock_sock(sk);

	err = kgrp_hci_sock_bound_ioctl(sk, cmd, arg);

done:
	release_sock(sk);
	return err;
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "kgr_patch_bsc1187054.h"
#include "../kallsyms_relocs.h"

#define KGR_PATCHED_MODULE "bluetooth"

static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "hci_get_dev_list", (void *)&kgre_hci_get_dev_list, "bluetooth" },
	{ "hci_dev_cmd", (void *)&kgre_hci_dev_cmd, "bluetooth" },
	{ "hci_dev_reset_stat", (void *)&kgre_hci_dev_reset_stat, "bluetooth" },
	{ "hci_dev_open", (void *)&kgre_hci_dev_open, "bluetooth" },
	{ "hci_dev_close", (void *)&kgre_hci_dev_close, "bluetooth" },
	{ "hci_dev_reset", (void *)&kgre_hci_dev_reset, "bluetooth" },
	{ "hci_get_dev_info", (void *)&kgre_hci_get_dev_info, "bluetooth" },
	{ "hci_get_conn_list", (void *)&kgre_hci_get_conn_list, "bluetooth" },
	{ "hci_get_conn_info", (void *)&kgre_hci_get_conn_info, "bluetooth" },
	{ "hci_get_auth_info", (void *)&kgre_hci_get_auth_info, "bluetooth" },
	{ "hci_inquiry", (void *)&kgre_hci_inquiry, "bluetooth" },
	{ "hci_bdaddr_list_add", (void *)&kgre_hci_bdaddr_list_add,
	  "bluetooth" },
	{ "hci_bdaddr_list_del", (void *)&kgre_hci_bdaddr_list_del,
	  "bluetooth" },
};

static int kgr_patch_bsc1187054_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1187054_module_nb = {
	.notifier_call = kgr_patch_bsc1187054_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1187054_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = __kgr_resolve_kallsyms_relocs(kgr_funcs,
						    ARRAY_SIZE(kgr_funcs));
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1187054_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1187054_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1187054_module_nb);
}

#endif /* IS_ENABLED(CONFIG_BT) */
